/*****************************************************************************
 * Author: Stephen Rhodes
 * Copyright 2018
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <assert.h>
#include <vlc_modules.h>
#include <vlc_services_discovery.h>
#include <vlc_input_item.h>
#include <vlc_interface.h>
#include <vlc_playlist.h>
#include <vlc_url.h>
#include <vlc_keystore.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#ifdef WIN32
	#include <winsock2.h>
	#include <ws2tcpip.h>
	#include <wincrypt.h>
	#include <iphlpapi.h>
	#include <fcntl.h>
#else
	#include <sys/socket.h>
	#include <arpa/inet.h>
	#include <unistd.h>
#endif

#include <nettle/base64.h>
#include <nettle/sha1.h>

//#include <libonvif/onvif.h>
#include "onvif.h"

char buf[128][8192];
int len[128];
char xaddrs_buf[128][1024];
int number_of_cameras;
static const int NO_VALUE = -3;
char uuid[47] = {0};
time_t time_offset;
xmlDocPtr xml_configs;

xmlDocPtr getSystemDateAndTime(char * xaddrs);
void getDiscoveryXml(char[], int);
void getDiscoveryXml2(char[], int);
bool getCameraName(xmlDocPtr, char[1024]);
xmlDocPtr getProfiles(char * username, char * password, char * xaddrs);
xmlDocPtr getStreamUri(char * token, char * username, char * password, char * xaddrs);
xmlDocPtr getDeviceInformation(char * username, char * password, char * xaddrs);
int setSocketOptions(int socket);
int getLastError(void);
void getScopeField(char *, char *, char[1024]);
void addCameraToPlaylist(char[], int, void *data);
int getXaddrsBufSize();
void clearBuffer();
void broadcast(void *data, int msg_id);
void readConfigFile();
int getCameraNameFromConfigs(char name_buf[], int buf_length, char * serial_number);
time_t getTimeOffset(char * xaddrs);
struct OnvifData onvif_data;

struct services_discovery_sys_t
{
  vlc_thread_t thread;
};
//} services_discovery_sys_t;

#define DEFAULT_USERNAME_TEXT  N_("Default User Name for IP Cameras")
#define DEFAULT_PASSWORD_TEXT  N_("Default Password for IP Cameras")
#define DEFAULT_USERNAME_LONGTEXT  N_("Default user name for IP cameras will be tried first before prompt")
#define DEFAULT_PASSWORD_LONGTEXT  N_("Default password for IP Cameras will be tried first before prompt")

static int Open( vlc_object_t * );
static void Close( vlc_object_t * );
static void *Run( void *data );

VLC_SD_PROBE_HELPER( "Onvif", N_("Onvif Camera Discovery"), SD_CAT_LAN )

vlc_module_begin()
    set_shortname( "Onvif" )
    set_description( N_( "Onvif Camera Discovery" ) )
    set_category( CAT_PLAYLIST )
		set_subcategory( SUBCAT_PLAYLIST_SD )
		add_string( "default-username", NULL, DEFAULT_USERNAME_TEXT, DEFAULT_USERNAME_LONGTEXT, true )
		add_string( "default-password", NULL, DEFAULT_PASSWORD_TEXT, DEFAULT_PASSWORD_LONGTEXT, true )
    set_capability( "services_discovery", 0 )
    set_callbacks( Open, Close )
    add_shortcut( "Onvif" )
    VLC_SD_PROBE_SUBMODULE

vlc_module_end()

static void *Run( void *data ) {
  services_discovery_t * p_sd = data;
	readConfigFile();

	#ifdef WIN32
	  WSADATA wsaData;
	  WSAStartup(MAKEWORD(2,2), &wsaData);
	#endif

	broadcast(p_sd, 1);

	int i;
  for (i=0; i<number_of_cameras; i++) {
		addCameraToPlaylist(buf[i], len[i], p_sd);
	}

	broadcast(p_sd, 2);
	for (i=0; i<number_of_cameras; i++) {
		addCameraToPlaylist(buf[i], len[i], p_sd);
	}

	#ifdef WIN32
		Sleep(1000);
	#else
		sleep(1);
	#endif
	broadcast(p_sd, 1);

	 for (i=0; i<number_of_cameras; i++) {
		addCameraToPlaylist(buf[i], len[i], p_sd);
	}

	broadcast(p_sd, 2);
	for (i=0; i<number_of_cameras; i++) {
		addCameraToPlaylist(buf[i], len[i], p_sd);
	}


	#ifdef WIN32
		Sleep(3000);
	#else
		sleep(3);
	#endif
	broadcast(p_sd, 1);

	 for (i=0; i<number_of_cameras; i++) {
		addCameraToPlaylist(buf[i], len[i], p_sd);
	}

	broadcast(p_sd, 2);
	for (i=0; i<number_of_cameras; i++) {
		addCameraToPlaylist(buf[i], len[i], p_sd);
	}

	#ifdef WIN32
	  WSACleanup();
	#endif

  return NULL;
}

static int Open( vlc_object_t *p_this )
{
    services_discovery_t *p_sd = (services_discovery_t *)p_this;
    struct services_discovery_sys_t *p_sys = NULL;

    p_sd->p_sys = p_sys = calloc(1, sizeof(services_discovery_sys_t));

    if (!p_sys) {
        return VLC_ENOMEM;
    }

    p_sd->description = _("Onvif Camera Discovery");

    if (vlc_clone (&p_sys->thread, Run, p_sd, VLC_THREAD_PRIORITY_LOW))
    {
        free (p_sys);
        return VLC_EGENERIC;
    }

    return VLC_SUCCESS;
}

static void Close( vlc_object_t *p_this )
{
    services_discovery_t *p_sd = (services_discovery_t *)p_this;
    struct services_discovery_sys_t *p_sys = p_sd->p_sys;

    free(p_sys);
}

void broadcast(void *data, int msg_id) {
	services_discovery_t * p_sd = data;
	struct sockaddr_in broadcast_address;
	int broadcast_socket;
	char broadcast_message[1024] = {0};
	unsigned int address_size;
	int error_code;

	if (msg_id == 1)
		getDiscoveryXml(broadcast_message, 1024);
	else if (msg_id == 2)
		getDiscoveryXml2(broadcast_message, 1024);
  int broadcast_message_length = strlen(broadcast_message);
  broadcast_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  setSocketOptions(broadcast_socket);
	clearBuffer();

  memset((char *) &broadcast_address, 0, sizeof(broadcast_address));
  broadcast_address.sin_family = AF_INET;
  broadcast_address.sin_port = htons(3702);
  broadcast_address.sin_addr.s_addr = inet_addr("239.255.255.250");
  int status = sendto(broadcast_socket, broadcast_message, broadcast_message_length, 0, (struct sockaddr*)&broadcast_address, sizeof(broadcast_address));
	if (status < 0) {
		error_code = getLastError();
		msg_Err(p_sd, "sendto error = %d\n", error_code);
	}

	int i = 0;
	unsigned char looping = 1;
	address_size = sizeof(broadcast_address);
	while(looping) {
    len[i] = recvfrom(broadcast_socket, buf[i], sizeof(buf[i]), 0, (struct sockaddr*) &broadcast_address, &address_size);
    if (len[i] > 0) {
      buf[i][len[i]] = '\0';
      i++;
    }
    else {
      looping = 0;
			if (len[i] < 0) {
				error_code = getLastError();
			}
    }
  }
	number_of_cameras = i;
	#ifdef WIN32
	  closesocket(broadcast_socket);
	#else
	  close(broadcast_socket);
	#endif
}

void addCameraToPlaylist(char buffer[], int buf_size, void *data) {
	services_discovery_t * p_sd = data;

	char input_username[128] = {'\0'};
	char input_password[128] = {'\0'};
	char *psz_default_username = var_CreateGetString( p_sd, "default-username" );
	char *psz_default_password = var_CreateGetString( p_sd, "default-password" );

	bool logging_in = true;
	bool duplicate = false;

	while (logging_in) {

		xmlDocPtr xml_input = xmlParseMemory(buffer, buf_size);
		char uuid_reply[128];
		if (getXmlValue(xml_input, (xmlChar *) "//s:Header//a:RelatesTo", uuid_reply, 128) == 0) {
			if (strcmp(uuid, uuid_reply) == 0) {
				//uuid match
			}
		}
		char camera_name[1024];
		bool vlc_name = getCameraName(xml_input, camera_name);
		char xaddrs[1024];

		if (strlen(camera_name)  == 0)
			strcpy(camera_name, "UNKNOWN CAMERA");

		if (getXmlValue(xml_input, (xmlChar *) "//s:Body//d:ProbeMatches//d:ProbeMatch//d:XAddrs", xaddrs, 1024) == 0) {
			//printf("xaddrs = %s\n", xaddrs);
			int xaddrs_buf_size = getXaddrsBufSize();
			for (int j=0; j<xaddrs_buf_size; j++) {
				if (strcmp((char*)xaddrs, xaddrs_buf[j]) == 0) {
					duplicate = true;
					logging_in = false;
				}
			}
			if (!duplicate) {
				sprintf(xaddrs_buf[xaddrs_buf_size], "%s", xaddrs);
				clearData(&onvif_data);
				char * sub = strstr(xaddrs, " ");
				if (sub != NULL) {
					int mark = sub - xaddrs;
					xaddrs[mark] = '\0';
				}
				strcpy(onvif_data.xaddrs, xaddrs);
				strcpy(onvif_data.device_service, onvif_data.xaddrs);
				extractOnvifService(onvif_data.device_service, true);

				time_offset = getTimeOffset(xaddrs);
				onvif_data.time_offset = time_offset;

				vlc_url_t camera_url;
				vlc_UrlParse(&camera_url, xaddrs);

				if (psz_default_username != NULL)
					strcpy(input_username, psz_default_username);
				if (psz_default_password != NULL)
					strcpy(input_password, psz_default_password);

				vlc_credential credential;
				vlc_credential_init( &credential, &camera_url );

				if (strlen(input_username) == 0 || strlen(input_password) == 0) {
					if( vlc_credential_get( &credential, p_sd, "rtsp-user", "rtsp-pwd", _("Onvif Camera Authentication"),  camera_name ) ) {
						strcpy(input_username, credential.psz_username);
						strcpy(input_password, credential.psz_password);
					}
					else {
						vlc_credential_clean(&credential);
						logging_in = false;
					}
				}
				else {
					credential.psz_username = (char*)malloc(128);
					credential.psz_password = (char*)malloc(128);
					strcpy(credential.psz_username, input_username);
					strcpy(credential.psz_password, input_password);
				}

				if (logging_in) {

					char * camera_username = (char*)malloc(128);
					char * camera_password = (char*)malloc(128);
					strcpy(camera_username, input_username);
					strcpy(camera_password, input_password);

					strcpy(onvif_data.username, camera_username);
					strcpy(onvif_data.password, camera_password);

					getCapabilities(&onvif_data);

					xmlDocPtr reply;
					reply = getProfiles(camera_username, camera_password, xaddrs);
					if (reply != NULL) {
						char error_msg[1024];
						bool auto_login_failure = false;
						if (getXmlValue(reply, (xmlChar *) "//s:Body//s:Fault//s:Code//s:Subcode//s:Value", error_msg, 1024) != NO_VALUE) {
							auto_login_failure = true;
						}

						while(auto_login_failure && logging_in) {
							if( vlc_credential_get( &credential, p_sd, "rtsp-user", "rtsp-pwd", _("Onvif Camera Authentication"),  camera_name ) ) {
								strcpy(camera_username, credential.psz_username);
								strcpy(camera_password, credential.psz_password);
								reply = getProfiles(camera_username, camera_password, xaddrs);
								if (getXmlValue(reply, (xmlChar *) "//s:Body//s:Fault//s:Code//s:Subcode//s:Value", error_msg, 1024) == NO_VALUE) {
									auto_login_failure = false;
								}
								else {
									//printf("reply = %s\n", reply);
								}
							}
							else {
								vlc_credential_clean(&credential);
								logging_in = false;
							}
						}

						if (!auto_login_failure) {
							char profile_token[128];
							if (getNodeAttribute(reply, (xmlChar *) "//s:Body//trt:GetProfilesResponse//trt:Profiles", "token", profile_token, 128) == 0) {
								char video_encoder_token[128];
								getNodeAttribute(reply, (xmlChar *) "//s:Body//trt:GetProfilesResponse//trt:Profiles//tt:VideoEncoderConfiguration", "token", video_encoder_token, 128);
								char video_source_token[128];
								getXmlValue(reply, (xmlChar *) "//s:Body//trt:GetProfilesResponse//trt:Profiles//tt:VideoSourceConfiguration//tt:SourceToken", video_source_token, 128);
								char uri[1024];
								reply = getStreamUri(profile_token, camera_username, camera_password, xaddrs);
								if (getXmlValue(reply, (xmlChar *) "//s:Body//trt:GetStreamUriResponse//trt:MediaUri//tt:Uri", uri, 1024) < 0) {
									getXmlValue(reply, (xmlChar *) "//s:Body//s:Fault//s:Code//s:Subcode//s:Value", error_msg, 1024);
									msg_Err(p_sd, "error_msg = %s\n", error_msg);
									return;
								}

								vlc_url_t url_uri;
								vlc_UrlParse(&url_uri, uri);
								credential.p_url = &url_uri;
								vlc_credential_store(&credential, p_sd);
								vlc_credential_clean(&credential);
								vlc_UrlClean (&url_uri);

								char serial_number[128];
								reply = getDeviceInformation(camera_username, camera_password, xaddrs);
								if (getXmlValue(reply, (xmlChar *) "//s:Body//tds:GetDeviceInformationResponse//tds:SerialNumber", serial_number, 128) < 0) {
									getXmlValue(reply, (xmlChar *) "//s:Body//s:Fault//s:Code//s:Subcode//s:Value", error_msg, 1024);
									msg_Err(p_sd, "error_msg = %s\n", error_msg);
									return;
								}
								char name_buf[128] = {0};
								if (getCameraNameFromConfigs(name_buf, 128, serial_number) == 0) {
									strcpy(camera_name, name_buf);
								}

								input_item_t *item = input_item_NewStream(uri, camera_name, -1);
								xmlDocPtr doc = xmlNewDoc((xmlChar*) "1.0");
								xmlNodePtr root = xmlNewDocNode(doc, NULL, (xmlChar*) "Parameters", NULL);
								xmlDocSetRootElement(doc, root);
								xmlNewTextChild(root, NULL, (xmlChar*) "XAddrs", (xmlChar*) xaddrs);

								input_item_SetAlbum(item, xaddrs);
								input_item_SetGenre(item, profile_token);
								xmlNewTextChild(root, NULL, (xmlChar*) "Profile", (xmlChar*) profile_token);
								input_item_SetCopyright(item, video_source_token);
								xmlNewTextChild(root, NULL, (xmlChar*) "VideoSourceToken", (xmlChar*) video_source_token);
								input_item_SetEncodedBy(item, video_encoder_token);
								xmlNewTextChild(root, NULL, (xmlChar*) "VideoEncoderToken", (xmlChar*) video_encoder_token);
								char time_offset_buf[128];
								sprintf(time_offset_buf, "%d", time_offset);
								input_item_SetLanguage(item, time_offset_buf);
								xmlNewTextChild(root, NULL, (xmlChar*) "TimeOffset", (xmlChar*) time_offset_buf);

								xmlChar* temp_buf;
								int temp_buf_size;
								xmlDocDumpFormatMemoryEnc(doc, &temp_buf, &temp_buf_size, "UTF-8", 1);
								input_item_SetDescription(item, (char*) serial_number);
								free(temp_buf);
								xmlCleanupParser();

								services_discovery_AddItem(p_sd, item);
							}
						}
					}
					vlc_UrlClean (&camera_url);
					free(camera_username);
					free(camera_password);
					xmlFreeDoc(xml_input);

					logging_in = false;
				}
				else {
					logging_in = false;
				}
			}
		}
	}
}


void getDiscoveryXml2(char buffer[], int buf_size) {
	char * string = "<s:Envelope xmlns:s=\"http://www.w3.org/2003/05/soap-envelope\" xmlns:a=\"http://schemas.xmlsoap.org/ws/2004/08/addressing\"><s:Header><a:Action s:mustUnderstand=\"1\">http://schemas.xmlsoap.org/ws/2005/04/discovery/Probe</a:Action><a:MessageID>uuid:6bbdae2d-f229-42c8-a27b-93880fb80826</a:MessageID><a:ReplyTo><a:Address>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</a:Address></a:ReplyTo><a:To s:mustUnderstand=\"1\">urn:schemas-xmlsoap-org:ws:2005:04:discovery</a:To></s:Header><s:Body><Probe xmlns=\"http://schemas.xmlsoap.org/ws/2005/04/discovery\"><d:Types xmlns:d=\"http://schemas.xmlsoap.org/ws/2005/04/discovery\" xmlns:dp0=\"http://www.onvif.org/ver10/device/wsdl\">dp0:Device</d:Types></Probe></s:Body></s:Envelope>";

	strcpy(buffer, string);
}



void getDiscoveryXml(char buffer[], int buf_size) {
	for (int i=0; i<buf_size; i++)
		buffer[i] = '\0';
	getUUID(uuid);
  xmlDocPtr doc = xmlNewDoc((xmlChar*) "1.0");
	xmlNodePtr root = xmlNewDocNode(doc, NULL, (xmlChar*) "Envelope", NULL);
	xmlDocSetRootElement(doc, root);
	xmlNewProp(root, (xmlChar*) "xmlns:SOAP-ENV", (xmlChar*) "http://www.w3.org/2003/05/soap-envelope");
	xmlNewProp(root, (xmlChar*) "xmlns:a", (xmlChar*) "http://schemas.xmlsoap.org/ws/2004/08/addressing");
	xmlNsPtr ns_env = xmlNewNs(root, NULL, (xmlChar*) "SOAP-ENV");
	xmlNsPtr ns_a = xmlNewNs(root, NULL, (xmlChar*) "a");
	xmlSetNs(root, ns_env);
  xmlNodePtr header = xmlNewTextChild(root, ns_env, (xmlChar*) "Header", NULL);
  xmlNodePtr action = xmlNewTextChild(header, ns_a, (xmlChar*) "Action", (xmlChar*) "http://schemas.xmlsoap.org/ws/2005/04/discovery/Probe");
  xmlNewProp(action, (xmlChar*) "SOAP-ENV:mustUnderstand", (xmlChar*) "1");
	xmlNodePtr messageid = xmlNewTextChild(header, ns_a, (xmlChar*) "MessageID", (xmlChar*) uuid);
  xmlNodePtr replyto = xmlNewTextChild(header, ns_a, (xmlChar*) "ReplyTo", NULL);
  xmlNodePtr address = xmlNewTextChild(replyto, ns_a, (xmlChar*) "Address", (xmlChar*) "http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous");
  xmlNodePtr to = xmlNewTextChild(header, ns_a, (xmlChar*) "To", (xmlChar*) "urn:schemas-xmlsoap-org:ws:2005:04:discovery");
  xmlNewProp(to, (xmlChar*) "SOAP-ENV:mustUnderstand", (xmlChar*) "1");
	xmlNodePtr body = xmlNewTextChild(root, ns_env, (xmlChar*) "Body", NULL);
  xmlNodePtr probe = xmlNewTextChild(body, NULL, (xmlChar*) "Probe", NULL);
  xmlNewProp(probe, (xmlChar*) "xmlns:p", (xmlChar*) "http://schemas.xmlsoap.org/ws/2005/04/discovery");
  xmlNsPtr ns_p = xmlNewNs(probe, NULL, (xmlChar*) "p");
  xmlSetNs(probe, ns_p);
	xmlNodePtr types = xmlNewTextChild(probe, NULL, (xmlChar*) "Types", (xmlChar*) "dp0:NetworkVideoTransmitter");
  xmlNewProp(types, (xmlChar*) "xmlns:d", (xmlChar*) "http://schemas.xmlsoap.org/ws/2005/04/discovery");
  xmlNewProp(types, (xmlChar*) "xmlns:dp0", (xmlChar*) "http://www.onvif.org/ver10/network/wsdl");
  xmlNsPtr ns_d = xmlNewNs(types, NULL, (xmlChar*) "d");
  xmlSetNs(types, ns_d);
  xmlOutputBufferPtr outputbuffer = xmlAllocOutputBuffer(NULL);
  xmlNodeDumpOutput(outputbuffer, doc, root, 0, 0, NULL);
  int size = xmlOutputBufferGetSize(outputbuffer);
  strcpy(buffer, (char*)xmlOutputBufferGetContent(outputbuffer));
  xmlOutputBufferFlush(outputbuffer);
  xmlOutputBufferClose(outputbuffer);
  xmlFreeDoc(doc);
}


xmlDocPtr getSystemDateAndTime(char * xaddrs) {
	xmlDocPtr doc = xmlNewDoc((xmlChar*) "1.0");
	xmlNodePtr root = xmlNewDocNode(doc, NULL, (xmlChar*) "Envelope", NULL);
	xmlDocSetRootElement(doc, root);
	xmlNsPtr ns_env = xmlNewNs(root, BAD_CAST "http://www.w3.org/2003/05/soap-envelope", BAD_CAST "SOAP-ENV");
  xmlNsPtr ns_tds = xmlNewNs(root, BAD_CAST "http://www.onvif.org/ver10/device/wsdl", BAD_CAST "tds");
	xmlSetNs(root, ns_env);
	xmlNodePtr body = xmlNewTextChild(root, ns_env, (xmlChar*) "Body", NULL);
	xmlNewTextChild(body, ns_tds, (xmlChar*) "GetSystemDateAndTime", NULL);
	char cmd[4096] = {0};
	addHttpHeader(doc, root, xaddrs, onvif_data.device_service, cmd, 4096);
	return sendCommandToCamera(cmd, xaddrs);
}

time_t getTimeOffset(char * xaddrs) {
	time_t result = 0;
	xmlDocPtr date_and_time = getSystemDateAndTime(xaddrs);
	if (date_and_time != NULL) {
		char hour_buf[16];
		char min_buf[16];
		char sec_buf[16];
		char year_buf[16];
		char month_buf[16];
		char day_buf[16];
		char dst_buf[16];
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:UTCDateTime//tt:Time//tt:Hour", hour_buf, 16);
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:UTCDateTime//tt:Time//tt:Minute", min_buf, 16);
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:UTCDateTime//tt:Time//tt:Second", sec_buf, 16);
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:UTCDateTime//tt:Date//tt:Year", year_buf, 16);
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:UTCDateTime//tt:Date//tt:Month", month_buf, 16);
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:UTCDateTime//tt:Date//tt:Day", day_buf, 16);
		getXmlValue(date_and_time, (xmlChar *) "//s:Body//tds:GetSystemDateAndTimeResponse//tds:SystemDateAndTime//tt:DaylightSavings", dst_buf, 16);

		int is_dst = 0;
		if (strcmp(dst_buf, "true") == 0)
			is_dst = 1;

		time_t now = time(NULL);
		struct tm *utc_here = gmtime(&now);
		utc_here->tm_isdst = -1;
		time_t utc_time_here = mktime(utc_here);

		struct tm *utc_there = localtime(&now);
		utc_there->tm_year = atoi(year_buf) - 1900;
		utc_there->tm_mon = atoi(month_buf) - 1;
		utc_there->tm_mday = atoi(day_buf);
		utc_there->tm_hour = atoi(hour_buf);
		utc_there->tm_min = atoi(min_buf);
		utc_there->tm_sec = atoi(sec_buf);
		utc_there->tm_isdst = is_dst;
		time_t utc_time_there = mktime(utc_there);

		result = utc_time_there - utc_time_here;
	}
	return result;
}

xmlDocPtr getDeviceInformation(char * username, char * password, char * xaddrs) {
  xmlDocPtr doc = xmlNewDoc((xmlChar*) "1.0");
  xmlNodePtr root = xmlNewDocNode(doc, NULL, (xmlChar*) "Envelope", NULL);
  xmlDocSetRootElement(doc, root);
	xmlNsPtr ns_env = xmlNewNs(root, BAD_CAST "http://www.w3.org/2003/05/soap-envelope", BAD_CAST "SOAP-ENV");
  xmlNsPtr ns_tds = xmlNewNs(root, BAD_CAST "http://www.onvif.org/ver10/device/wsdl", BAD_CAST "tds");
  xmlSetNs(root, ns_env);
  addUsernameDigestHeader(root, ns_env, username, password, time_offset);
  xmlNodePtr body = xmlNewTextChild(root, ns_env, (xmlChar*) "Body", NULL);
  xmlNewTextChild(body, ns_tds, (xmlChar*) "GetDeviceInformation", NULL);
  char cmd[4096] = {0};
  addHttpHeader(doc, root, xaddrs, onvif_data.device_service, cmd, 4096);
  return sendCommandToCamera(cmd, xaddrs);
}

xmlDocPtr getStreamUri(char * token, char * username, char * password, char * xaddrs) {
	xmlDocPtr doc = xmlNewDoc((xmlChar*) "1.0");
	xmlNodePtr root = xmlNewDocNode(doc, NULL, (xmlChar*) "Envelope", NULL);
	xmlDocSetRootElement(doc, root);
	xmlNsPtr ns_env = xmlNewNs(root, BAD_CAST "http://www.w3.org/2003/05/soap-envelope", BAD_CAST "SOAP-ENV");
  xmlNsPtr ns_trt = xmlNewNs(root, BAD_CAST "http://www.onvif.org/ver10/media/wsdl", BAD_CAST "trt");
	xmlNsPtr ns_tt = xmlNewNs(root, BAD_CAST "http://www.onvif.org/ver10/schema", BAD_CAST "tt");
	xmlSetNs(root, ns_env);
	addUsernameDigestHeader(root, ns_env, username, password, time_offset);
	xmlNodePtr body = xmlNewTextChild(root, ns_env, (xmlChar*) "Body", NULL);
	xmlNodePtr getStreamUri = xmlNewTextChild(body, ns_trt, (xmlChar*) "GetStreamUri", NULL);
	xmlNodePtr streamSetup = xmlNewTextChild(getStreamUri, ns_trt, (xmlChar*) "StreamSetup", NULL);
	xmlNewTextChild(streamSetup, ns_tt, (xmlChar*) "Stream", (xmlChar*) "RTP-Unicast");
	xmlNodePtr transport = xmlNewTextChild(streamSetup, ns_tt, (xmlChar*) "Transport", NULL);
	xmlNewTextChild(transport, ns_tt, (xmlChar*) "Protocol", (xmlChar*) "RTSP");
	xmlNewTextChild(getStreamUri, ns_trt, (xmlChar*) "ProfileToken", (xmlChar*) token);
	char cmd[4096] = {0};
	addHttpHeader(doc, root, xaddrs, onvif_data.media_service, cmd, 4096);
	return sendCommandToCamera(cmd, xaddrs);
}

xmlDocPtr getProfiles(char * username, char * password, char * xaddrs) {
	xmlDocPtr doc = xmlNewDoc((xmlChar*) "1.0");
	xmlNodePtr root = xmlNewDocNode(doc, NULL, (xmlChar*) "Envelope", NULL);
	xmlDocSetRootElement(doc, root);
	xmlNsPtr ns_env = xmlNewNs(root, BAD_CAST "http://www.w3.org/2003/05/soap-envelope", BAD_CAST "SOAP-ENV");
	xmlNsPtr ns_trt = xmlNewNs(root, BAD_CAST "http://www.onvif.org/ver10/media/wsdl", BAD_CAST "trt");
	xmlSetNs(root, ns_env);
	addUsernameDigestHeader(root, ns_env, username, password, time_offset);
	xmlNodePtr body = xmlNewTextChild(root, ns_env, (xmlChar*) "Body", NULL);
	xmlNewTextChild(body, ns_trt, (xmlChar*) "GetProfiles", NULL);
	char cmd[4096] = {0};
	addHttpHeader(doc, root, xaddrs, onvif_data.media_service, cmd, 4096);
	return sendCommandToCamera(cmd, xaddrs);
}

void readConfigFile() {
	char dirname[1024] = {0};

	char *psz_dir = config_GetUserDir( VLC_CONFIG_DIR );
	strcat(dirname, psz_dir);

	#ifdef WIN32
		strcat(dirname, "\\cameras.xml");
	#else
		strcat(dirname, "/cameras.xml");
	#endif

	xml_configs = xmlParseFile(dirname);
}

int getCameraNameFromConfigs(char name_buf[], int buf_length, char * arg) {
	int result = -1;

	if (xml_configs != NULL) {
		bool found = false;
		xmlChar * xpath = (xmlChar*) "//Cameras//Parameters";
		xmlXPathObjectPtr path_object = getNodeSet(xml_configs, xpath);
		xmlNodeSetPtr nodeset = path_object->nodesetval;
		for (int i=0; i<nodeset->nodeNr; i++) {
			xmlNodePtr cur = nodeset->nodeTab[i];
			xmlChar * serial_number = xmlGetProp(cur, (xmlChar*) "SerialNumber");
			if (strcmp((char*) serial_number, arg) == 0) {
				found = true;
				cur = cur->xmlChildrenNode;
				while (cur != NULL) {
					if (strcmp((char*) cur->name, "VLCName") == 0) {
						xmlChar * xml_name = xmlNodeGetContent(cur);
						if (strlen((char*) xml_name) < buf_length) {
							result = 0;
							strcpy(name_buf, (char*) xml_name);
						}
					}
					cur = cur->next;
				}
			}
		}
	}
	return result;
}

bool getCameraName(xmlDocPtr doc, char camera_name[1024]) {
	bool result = false;
	for(int i=0; i<1024; i++)
		camera_name[i] = '\0';

	char scopes[8192];
	getXmlValue(doc, "//s:Body//d:ProbeMatches//d:ProbeMatch//d:Scopes", scopes, 8192);

	char temp_mfgr[1024] = {0};
	char temp_hdwr[1024] = {0};

	getScopeField(scopes, "onvif://www.onvif.org/name/", temp_mfgr);
	getScopeField(scopes, "onvif://www.onvif.org/hardware/", temp_hdwr);

	if (strlen(temp_mfgr) > 0) {
		strcat(camera_name, temp_mfgr);
	}
	if (strlen(temp_hdwr) > 0) {
		if (strstr(temp_mfgr, temp_hdwr) == NULL) {
			strcat(camera_name, " ");
			strcat(camera_name, temp_hdwr);
		}
	}
	return result;
}

void getScopeField(char * scope, char * field_name, char cleaned[1024]) {
	char * field;
	char field_contents[1024] = {0};
	char * mark;
	int length;
	char * result = NULL;

	field = strstr(scope, field_name);
	if (field != NULL) {
		field = field + strlen(field_name);
		mark = strstr(field, " ");
		if (mark != NULL) {
			length = mark - field;
			strncpy(field_contents, field, length);
		}
		else {
			strcpy(field_contents, field);
		}

		length = strlen(field_contents);
		int offset = 0;
		int j;
		for (int i=0; i<length; i++) {
			j = i - offset;
			if (field_contents[i] == '%') {
				char middle[3] = {0};
				i++; offset++;
				middle[0] = field_contents[i];
				i++; offset++;
				middle[1] = field_contents[i];
				char * ptr;
				int result = strtol(middle, &ptr, 16);
				cleaned[j] = result;
			}
			else {
				cleaned[j] = field_contents[i];
			}
		}
		cleaned[length] = '\0';
	}
}

int setSocketOptions(int socket) {
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 500000;
	int broadcast = 500;
	char loopch = 0;
	int status = 0;
	struct in_addr localInterface;

	#ifdef WIN32

		PMIB_IPADDRTABLE pIPAddrTable;
		DWORD dwSize = 0;
		DWORD dwRetVal = 0;
		IN_ADDR IPAddr;
		LPVOID lpMsgBuf;

		pIPAddrTable = (MIB_IPADDRTABLE *) malloc(sizeof(MIB_IPADDRTABLE));
		if (pIPAddrTable) {
			if (GetIpAddrTable(pIPAddrTable, &dwSize, 0) == ERROR_INSUFFICIENT_BUFFER) {
				free(pIPAddrTable);
				pIPAddrTable = (MIB_IPADDRTABLE *) malloc(dwSize);
			}
			if (pIPAddrTable == NULL) {
				printf("Memory allocation failed for GetIpAddrTable\n");
				return -1;
			}
		}

		if ((dwRetVal = GetIpAddrTable(pIPAddrTable, &dwSize, 0)) != NO_ERROR) {
			printf("GetIpAddrTable failed with error %d\n", dwRetVal);
			return -1;
		}

		int p = 0;
		while (p < (int)pIPAddrTable->dwNumEntries) {
			IPAddr.S_un.S_addr = (u_long)pIPAddrTable->table[p].dwAddr;
			IPAddr.S_un.S_addr = (u_long)pIPAddrTable->table[p].dwMask;
			if (pIPAddrTable->table[p].dwAddr != inet_addr("127.0.0.1") && pIPAddrTable->table[p].dwMask == inet_addr("255.255.255.0")) {
				localInterface.s_addr = pIPAddrTable->table[p].dwAddr;
				status = setsockopt(socket, IPPROTO_IP, IP_MULTICAST_IF, (const char *)&localInterface, sizeof(localInterface));
				if (status < 0)
					printf("ip_multicast_if error");
				p = (int)pIPAddrTable->dwNumEntries;
			}
			p++;
		}

		if (pIPAddrTable) {
			free(pIPAddrTable);
			pIPAddrTable = NULL;
		}

		status = setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, (const char *)&broadcast, sizeof(broadcast));
	#else
		status = setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, (struct timeval *)&tv, sizeof(struct timeval));
	#endif
	status = setsockopt(socket, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loopch, sizeof(loopch));
	return 0;
}

int getLastError() {
	/*
	#ifdef WIN32
		return WSAGetLastError();
	#else
		return errno;
	#endif
	*/
	return 0;
}

int getXaddrsBufSize() {
	int size = 0;
	bool found = false;
	while (!found) {
		if (strlen(xaddrs_buf[size]) == 0) {
			found = true;
		}
		else {
			size++;
			if (size > 127)
				found = true;
		}
	}
	return size;
}

void clearBuffer() {
	for (int i=0; i<128; i++) {
		for (int j=0; j<8192; j++) {
			buf[i][j] = '\0';
		}
	}
}
